# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import glob
import struct


def generate_fname(nout, path="", ftype="", cpuid=1, ext=""):

    if len(path) > 0:
        if path[-1] != "/":
            path = path + "/"

    if nout == -1:
        filelist = sorted(glob.glob(path + "output*"))
        number = filelist[-1].split("_")[-1]
    else:
        number = str(nout).zfill(5)

    infile = path + "output_" + number
    if len(ftype) > 0:
        infile += "/" + ftype + "_" + number
        if cpuid >= 0:
            infile += ".out" + str(cpuid).zfill(5)

    if len(ext) > 0:
        infile += ext

    return infile


def read_parameter_file(fname=None, delimiter="="):
    """
    Read info file and create dictionary
    """
    out = {}
    with open(fname) as f:
        content = f.readlines()
    for line in content:
        sp = line.split(delimiter)
        if len(sp) > 1:
            value = sp[1].strip()
            try:
                value = eval(value)
            except NameError:
                pass
            out[sp[0].strip()] = value
    return out


def read_binary_data(content=None,
                     fmt=None,
                     offsets=None,
                     skip_head=True,
                     increment=True):
    """
    Unpack binary data from a content buffer using a dict of offsets.
    Also increment the offsets of the corresponding data read, as well as
    increase the line count by 1.
    """

    byte_size = {
        "b": 1,
        "h": 2,
        "i": 4,
        "q": 8,
        "f": 4,
        "d": 8,
        "e": 8,
        "n": 8,
        "l": 8,
        "s": 1
    }

    offset = 0
    for key in offsets:
        offset += offsets[key] * byte_size[key]
    # if offset is None:
    #     offset = 4*ninteg + 8*(nlines+nfloat+nlongi) + nstrin + nquadr*16
    if skip_head:
        offset += 4  # + correction

    # byte_size = {"b": 1 , "h": 2, "i": 4, "q": 8, "f": 4, "d": 8, "e": 8}
    if len(fmt) == 1:
        mult = 1
    else:
        mult = int(fmt[:-1])
    pack_size = mult * byte_size[fmt[-1]]

    if increment:
        offsets[fmt[-1]] += mult
    offsets["n"] += 1

    return struct.unpack(fmt, content[offset:offset + pack_size])



import numpy as np
from .. import units


def get_finite_inds(x):
    """
    Find indices of finite numbers.
    """
    return np.where(np.isfinite(x))[0]


def finmin(x):
    """
    Finite minimum.
    """
    return np.amin(x.take(get_finite_inds(x)))


def finmax(x):
    """
    Finite maximum.
    """
    return np.amax(x.take(get_finite_inds(x)))


def to_bin_centers(x):
    """
    Convert array edges to centers
    """
    return 0.5 * (x[1:] + x[:-1])


def to_bin_edges(x):
    """
    Convert array centers to edges
    """
    centers = to_bin_centers(x)
    left = centers[0] - (x[1] - x[0])
    right = centers[-1] + (x[-1] - x[-2])
    return np.concatenate(np.concatenate(left, centers), right)


def perpendicular_vector(v):
    """
    Compute a vector perpendicular to the input vector
    """

    # x = y = z = 0 is not an acceptable solution
    if v[0] == v[1] == v[2] == 0:
        raise ValueError("zero-vector")

    if v[2] == 0:
        return [-v[1], v[0], 0]
    else:
        return [1.0, 1.0, -1.0 * (v[0] + v[1]) / v[2]]


def value_to_string(val, precision=3):
    if (not isinstance(val, float)) or (val == 0):
        text = str(val)
    elif (abs(val) >= 10.0**(precision)) or \
         (abs(val) <= 10.0**(-precision)):
        text = "{val:.{prec}e}".format(val=val, prec=precision)
    else:
        text = "{}".format(val)
        if len(text) > precision + 2 + (text[0] == '-'):
            text = "{val:.{prec}f}".format(val=val, prec=precision)
    return text


def make_label(name=None, unit=None):
    lab = ""
    if name:
        lab += name
    if unit and unit != units.dimensionless:
        if name:
            lab += " "
        lab += "[{:~}]".format(unit)
    return lab.strip()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from pint.quantity import Quantity
from .slice import get_slice_direction
from .render import render
from .parser import parse_layer
from ..core import Plot, Array


def plane(*layers,
          direction="z",
          dx=0.0,
          dy=0.0,
          fname=None,
          title=None,
          sinks=True,
          plot=True,
          mode=None,
          norm=None,
          vmin=None,
          vmax=None,
          operation="mean",
          origin=[0, 0, 0],
          resolution=256,
          ax=None,
          **kwargs):
    """
    Plot a 2D slice through the data domain.
    """

    if isinstance(layers, Array):
        layers = [layers]

    to_process = []
    to_render = []
    operations = []
    for layer in layers:
        data, settings, params = parse_layer(layer,
                                             mode=mode,
                                             norm=norm,
                                             vmin=vmin,
                                             vmax=vmax,
                                             operation=operation,
                                             **kwargs)

        to_process.append(data)
        to_render.append({
            "mode": settings["mode"],
            "params": params,
            "unit": data.unit.units,
            "name": data.name
        })
        operations.append(settings["operation"])

    dataset = to_process[0].parent

    # Set window size
    if dy == 0.0:
        dy = dx
    if not isinstance(dx, Quantity):
        dx *= dataset["xyz"].unit
    if not isinstance(dy, Quantity):
        dy *= dataset["xyz"].unit

    dir_vecs, origin = get_slice_direction(direction=direction,
                                           dataset=dataset,
                                           dx=0.5 * (dx + dy),
                                           origin=origin)

    # Distance to the plane
    xyz = dataset["xyz"] - origin
    diagonal = dataset["dx"] * np.sqrt(3.0) * 0.5
    dist1 = np.sum(xyz * dir_vecs[0],
                   axis=1)  # / np.linalg.norm(dir_vecs[0][1])

    # Distance from center
    dist2 = xyz - diagonal

    # Select only the cells in contact with the slice,
    # at a distance less than dx/2
    cube = np.ravel(
        np.where(
            np.logical_and(
                np.abs(dist1) <= 1.0001 * diagonal,
                np.abs(dist2.norm) <=
                max(dx.magnitude, dy.magnitude) * 0.5 * np.sqrt(2.0))))

    ncells = len(cube)

    # Project coordinates onto the plane by taking dot product with axes
    # vectors
    coords = xyz[cube]
    datax = np.inner(coords, dir_vecs[1])
    datay = np.inner(coords, dir_vecs[2])
    datadx = diagonal[cube]

    # Define slice extent and resolution
    xmin = -0.5 * dx
    xmax = xmin + dx
    ymin = -0.5 * dy
    ymax = ymin + dy
    nx = resolution
    ny = resolution
    dpx = (xmax - xmin) / float(nx)
    dpy = (ymax - ymin) / float(ny)
    x = np.linspace(xmin + 0.5 * dpx, xmax - 0.5 * dpx, nx).magnitude
    y = np.linspace(ymin + 0.5 * dpy, ymax - 0.5 * dpy, ny).magnitude

    counts = np.zeros([ny, nx])
    for ind in range(len(to_process)):
        if to_render[ind]["mode"] in ["vec", "stream"]:
            if to_process[ind].ndim < 3:
                to_process[ind] = to_process[ind].array[cube]
            else:
                uv = np.inner(to_process[ind].array.take(cube, axis=0),
                              dir_vecs[1:])
                w = None
                if "color" in to_render[ind]["params"]:
                    if isinstance(to_render[ind]["params"]["color"], Array):
                        w = to_render[ind]["params"]["color"].norm
                    elif isinstance(to_render[ind]["params"]["color"],
                                    np.ndarray):
                        w = to_render[ind]["params"]["color"]
                if w is None:
                    w = np.linalg.norm(uv, axis=1)
                else:
                    w = w.take(cube, axis=0)
                to_process[ind] = np.concatenate((uv, w.reshape(ncells, 1)),
                                                 axis=1)
            to_render[ind]["data"] = np.zeros(
                [ny, nx, to_process[ind].shape[1]])
        else:
            to_process[ind] = to_process[ind].array[cube]
            to_render[ind]["data"] = np.zeros([ny, nx])

    datax -= xmin
    datay -= ymin
    istart = ((datax - datadx) / dpx).array.astype(np.int64)
    iend = ((datax + datadx) / dpx).array.astype(np.int64) + 1
    jstart = ((datay - datadx) / dpy).array.astype(np.int64)
    jend = ((datay + datadx) / dpy).array.astype(np.int64) + 1

    for i in range(len(istart)):
        i0 = istart[i]
        i1 = iend[i]
        j0 = jstart[i]
        j1 = jend[i]
        if i0 <= nx and j0 <= ny and i1 > 0 and j1 > 0:
            i0 = max(i0, 0)
            i1 = min(i1, nx)
            j0 = max(j0, 0)
            j1 = min(j1, ny)
            for ind in range(len(to_process)):
                to_render[ind]["data"][j0:j1, i0:i1] += to_process[ind][i]
            counts[j0:j1, i0:i1] += 1.0

    # Normalize by counts
    mask = counts == 0.0
    for ind in range(len(to_process)):
        if to_render[ind]["data"].ndim > counts.ndim:
            to_render[ind]["data"] = np.ma.masked_where(
                np.broadcast_to(mask.reshape(ny, nx, 1),
                                to_render[ind]["data"].shape),
                to_render[ind]["data"]) / counts.reshape(ny, nx, 1)
        else:
            to_render[ind]["data"] = np.ma.masked_where(
                mask, to_render[ind]["data"]) / counts

    to_return = {"x": x, "y": y, "layers": to_render}
    if plot:
        # Render the map
        figure = render(x=x, y=y, data=to_render, ax=ax)
        figure["ax"].set_xlabel(dataset["xyz"].x.label)
        figure["ax"].set_ylabel(dataset["xyz"].y.label)
        if ax is None:
            figure["ax"].set_aspect("equal")
        to_return.update({"fig": figure["fig"], "ax": figure["ax"]})

    return Plot(**to_return)
